import sys
import time
from contextlib import contextmanager, redirect_stdout, redirect_stderr


class OutputBuffer:
    flush_length = 1000
    flush_time = 1

    def __init__(self, flush):
        self._flush = flush
        self.reset()

    def reset(self):
        self.parts = []
        self.last_time = time.time()

    def put(self, output_type, text, **extra):
        if isinstance(text, bytes):
            text = text.decode("utf8", "replace")
        if not isinstance(text, str):
            raise TypeError(f"Can only write str, not {type(text).__name__}")
        assert isinstance(output_type, str)

        if not self.parts or self.parts[-1]["type"] != output_type:
            self.parts.append(dict(type=output_type, text=text, **extra))
        else:
            self.parts[-1]["text"] += text

        if self.should_flush():
            self.flush()

    def should_flush(self):
        return (
            len(self.parts) > 1
            or self.last_time and time.time() - self.last_time > self.flush_time
            or sum(len(p["text"]) for p in self.parts) >= self.flush_length
        )

    def flush(self):
        if not self.parts:
            return
        self._flush(self.parts)
        self.reset()

    @contextmanager
    def redirect_std_streams(self):
        with (
            redirect_stdout(SysStream("stdout", self)),  # noqa
            redirect_stderr(SysStream("stderr", self)),  # noqa
        ):
            yield


class SysStream:
    def __init__(self, output_type, output_buffer):
        self.type = output_type
        self.output_buffer = output_buffer

    def __getattr__(self, item):
        return getattr(sys.__stdout__, item)

    def write(self, s):
        self.output_buffer.put(self.type, s)

    def flush(self):
        self.output_buffer.flush()



import ast
import builtins
import linecache
import logging
import sys
import time
from code import InteractiveConsole
from collections.abc import Awaitable
from contextlib import contextmanager
from types import ModuleType

from .output import OutputBuffer

log = logging.getLogger(__name__)


class Runner:
    OutputBufferClass = OutputBuffer

    def __init__(
        self,
        *,
        callback=None,
        filename="my_program.py",
    ):
        self.set_callback(callback)
        self.filename = filename

        self.console = InteractiveConsole()
        self.output_buffer = self.OutputBufferClass(
            lambda parts: self.callback("output", parts=parts)
        )
        self.reset()

    def set_callback(self, callback):
        self._callback = callback

    def callback(self, event_type, **data):
        if event_type != "output":
            self.output_buffer.flush()

        return self._callback(event_type, data)

    def output(self, output_type, text, **extra):
        return self.output_buffer.put(output_type, text, **extra)

    def execute(self, code_obj, source_code, mode=None):  # noqa
        return eval(code_obj, self.console.locals)  # noqa

    @contextmanager
    def _execute_context(self, source_code):
        with self.output_buffer.redirect_std_streams():
            try:
                yield
            except KeyboardInterrupt:
                raise
            except BaseException as e:
                self.output("traceback", **self.serialize_traceback(e, source_code))
        self.post_run()

    def run(self, source_code, mode="exec"):
        code_obj = self.pre_run(source_code, mode)
        with self._execute_context(source_code):
            if code_obj:
                return self.execute(code_obj, source_code, mode)

    async def run_async(self, source_code, mode="exec", top_level_await=True):
        code_obj = self.pre_run(source_code, mode, top_level_await=top_level_await)
        with self._execute_context(source_code):
            if code_obj:
                result = self.execute(code_obj, source_code, mode)
                while isinstance(result, Awaitable):
                    result = await result
                return result

    def serialize_traceback(self, exc, source_code):  # noqa
        raise NotImplementedError  # pragma: no cover

    def serialize_syntax_error(self, exc, source_code):
        return self.serialize_traceback(exc, source_code)

    def pre_run(self, source_code, mode="exec", top_level_await=False):
        compile_mode = mode
        if mode == "single":
            source_code += "\n"  # Allow compiling single-line compound statements
        elif mode != "eval":
            compile_mode = "exec"
            self.reset()
        self.output_buffer.reset()

        filename = self.filename
        linecache.cache[filename] = (
            len(source_code),
            0,
            [line + "\n" for line in source_code.splitlines()],
            filename,
        )

        try:
            return compile(
                source_code,
                filename,
                compile_mode,
                flags=top_level_await * ast.PyCF_ALLOW_TOP_LEVEL_AWAIT,
            )
        except SyntaxError as e:
            try:
                if not ast.parse(source_code).body:
                    # Code is only comments, which cannot be compiled in 'single' mode
                    return
            except SyntaxError:
                pass

            self.output("syntax_error", **self.serialize_syntax_error(e, source_code))

    def post_run(self):
        self.output_buffer.flush()

    def reset(self):
        mod = ModuleType("__main__")
        mod.__file__ = self.filename
        sys.modules["__main__"] = mod
        self.console.locals = mod.__dict__
        self.output_buffer.reset()


class PatchedStdinRunner(Runner):  # noqa
    def execute(self, *args, **kwargs):
        sys.stdin.readline = self.readline
        builtins.input = self.input
        return super().execute(*args, **kwargs)

    def reset(self):
        super().reset()
        self.line = ""

    def non_str_input(self, value):
        raise TypeError(f"Callback for input should return str, not {type(value).__name__}")

    def readline(self, n=-1, prompt=""):
        if not self.line and n:
            value = self.callback("input", prompt=prompt)
            if not isinstance(value, str):
                value = self.non_str_input(value) or ""
            if not value.endswith("\n"):
                value += "\n"
            self.output("input", value)
            self.line = value

        if n < 0 or n > len(self.line):
            n = len(self.line)
        to_return = self.line[:n]
        self.line = self.line[n:]
        return to_return

    def input(self, prompt=""):
        self.output("input_prompt", prompt)
        return self.readline(prompt=prompt)[:-1]  # Remove trailing newline


class PatchedSleepRunner(Runner):  # noqa
    def execute(self, *args, **kwargs):
        time.sleep = self.sleep
        return super().execute(*args, **kwargs)

    def sleep(self, seconds):
        if not isinstance(seconds, (int, float)):
            raise TypeError(f"an integer is required (got type {type(seconds).__name__})")
        if not seconds >= 0:
            raise ValueError("sleep length must be non-negative")
        return self.callback("sleep", seconds=seconds)


class PyodideRunner(PatchedStdinRunner, PatchedSleepRunner):  # noqa # pragma: no cover
    ServiceWorkerError = RuntimeError(
        "The service worker for reading input isn't working. "
        "Try closing all this site's tabs, then reopening. "
        "If that doesn't work, try using a different browser."
    )

    NoChannelError = RuntimeError(
        "This browser doesn't support reading input. "
        "Try upgrading to the most recent version or switching to a different browser, "
        "e.g. Chrome or Firefox."
    )
    InterruptError = KeyboardInterrupt

    def readline(self, *args, **kwargs):
        import pyodide  # noqa

        try:
            return super().readline(*args, **kwargs)
        except pyodide.JsException as e:
            typ = getattr(e.js_error, "type", "")
            err = getattr(self, typ, None)
            if err:
                raise err from None
            else:
                raise



import asyncio
import builtins
import time
import traceback
from textwrap import dedent

import pytest

from python_runner import PatchedStdinRunner, PatchedSleepRunner
from python_runner.output import OutputBuffer

OutputBuffer.flush_time = 0.1


class MyRunner(PatchedStdinRunner):
    def serialize_traceback(self, exc, source_code):
        return {
            "text": "".join(traceback.format_exception_only(type(exc), exc)),
            "source_code": source_code,
        }


events = []


def default_callback(event_type, data):
    events.append((event_type, data))
    if event_type == "input":
        return f"input: {len(events)}"


def check_simple(source_code, expected_events, mode="exec", runner=None):
    global events
    events = []

    runner = runner or MyRunner(callback=default_callback)
    result = runner.run(source_code, mode=mode)
    assert events == expected_events
    if mode != "eval":
        assert result is None
    return result


def test_simple_print():
    check_simple(
        "print(1); print(2)",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "stdout",
                            "text": "1\n2\n",
                        },
                    ],
                },
            ),
        ],
    )


def test_stdout_bytes():
    check_simple(
        "import sys; sys.stdout.write(b'abc' + '☃'.encode())",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "stdout",
                            "text": "abc☃",
                        },
                    ],
                },
            ),
        ],
    )


def test_stdout_write_non_str():
    check_simple(
        "import sys; sys.stdout.write(123)",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "traceback",
                            "text": "TypeError: Can only write str, not int\n",
                            "source_code": "import sys; sys.stdout.write(123)",
                        }
                    ],
                },
            ),
        ],
    )


def test_stdout_attrs():
    check_simple(
        "import sys; print(sys.stdout.encoding, callable(sys.stdout.isatty))",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "stdout",
                            "text": "utf-8 True\n",
                        },
                    ],
                },
            ),
        ],
    )


def test_mixed_output():
    source = dedent(
        """
        import sys
        
        print(1)
        print(2)
        
        print(3, file=sys.stderr)
        print(4, file=sys.stderr)
        
        print(5)
        print(6)
        
        print(7, file=sys.stderr)
        print(8, file=sys.stderr)
        
        1/0
        """
    )
    check_simple(
        source,
        [
            (
                "output",
                {
                    "parts": [
                        {"type": "stdout", "text": "1\n2\n"},
                        {"type": "stderr", "text": "3"},
                    ]
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {"type": "stderr", "text": "\n4\n"},
                        {"type": "stdout", "text": "5"},
                    ]
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {"type": "stdout", "text": "\n6\n"},
                        {"type": "stderr", "text": "7"},
                    ]
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "stderr",
                            "text": "\n8\n",
                        },
                        {
                            "type": "traceback",
                            "text": "ZeroDivisionError: division by zero\n",
                            "source_code": source,
                        },
                    ]
                },
            ),
        ],
    )


def test_syntax_error():
    check_simple(
        "a b",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "syntax_error",
                            "text": (
                                '  File "my_program.py", line 1\n'
                                "    a b\n"
                                "      ^\n"
                                "SyntaxError: invalid syntax\n"
                            ),
                            "source_code": "a b",
                        }
                    ]
                },
            )
        ],
    )


def test_runtime_error():
    check_simple(
        "nonexistent",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "traceback",
                            "text": "NameError: name 'nonexistent' is not defined\n",
                            "source_code": "nonexistent",
                        }
                    ]
                },
            )
        ],
    )


def test_simple_input():
    check_simple(
        "input('some_prompt'); print('after')",
        [
            (
                "output",
                {
                    "parts": [
                        {"type": "input_prompt", "text": "some_prompt"},
                    ]
                },
            ),
            (
                "input",
                {"prompt": "some_prompt"},
            ),
            (
                "output",
                {
                    "parts": [
                        {"type": "input", "text": "input: 2\n"},
                        {"type": "stdout", "text": "after"},
                    ]
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {"type": "stdout", "text": "\n"},
                    ]
                },
            ),
        ],
    )


def test_non_str_input():
    def callback(event_type, data):
        if event_type == "input":
            return
        return default_callback(event_type, data)

    check_simple(
        "input()",
        runner=MyRunner(callback=callback),
        expected_events=[
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "input_prompt",
                            "text": "",
                        },
                    ],
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "traceback",
                            "text": "TypeError: Callback for input should return str, not NoneType\n",
                            "source_code": "input()",
                        }
                    ]
                },
            ),
        ],
    )


def test_custom_non_str_input():
    class Runner(MyRunner):
        def non_str_input(self, value):
            print(repr(value))

    def callback(event_type, data):
        if event_type == "input":
            return 123
        return default_callback(event_type, data)

    runner = Runner(callback=callback)
    assert runner.line == ""

    check_simple(
        "input()",
        runner=runner,
        expected_events=[
            (
                "output",
                {
                    "parts": [
                        {"text": "", "type": "input_prompt"},
                    ],
                },
            ),
            (
                "output",
                {
                    "parts": [
                        {"text": "123\n", "type": "stdout"},
                        {"text": "\n", "type": "input"},
                    ]
                },
            ),
        ],
    )

    assert runner.line == ""


def test_single():
    check_simple(
        "1 + 2",
        [
            (
                "output",
                {
                    "parts": [
                        {"type": "stdout", "text": "3\n"},
                    ],
                },
            ),
        ],
        mode="single",
    )

    check_simple(
        "for i in range(2): print(i)",
        [
            (
                "output",
                {
                    "parts": [
                        {"type": "stdout", "text": "0\n1\n"},
                    ],
                },
            ),
        ],
        mode="single",
    )


def test_eval():
    assert 3 == check_simple(
        "1 + 2",
        [],
        mode="eval",
    )


def test_empty():
    for mode in ["single", "eval", "exec"]:
        for source in ["", "#", "#foo", "#foo\n#bar\n", "\n", " "]:
            check_simple(source, [], mode=mode)


def test_flush_direct():
    check_simple(
        dedent(
            """
            import sys
            
            print(1)
            print(2)
            
            sys.stdout.flush()
            sys.stdout.flush()
            
            print(3)
            print(4)
            """
        ),
        [
            ("output", {"parts": [{"type": "stdout", "text": "1\n2\n"}]}),
            ("output", {"parts": [{"type": "stdout", "text": "3\n4\n"}]}),
        ],
    )


def test_flush_time():
    check_simple(
        dedent(
            """
            import time

            print(1)
            print(2)

            time.sleep(0.11)

            print(3)
            print(4)
            """
        ),
        [
            ("output", {"parts": [{"type": "stdout", "text": "1\n2\n3"}]}),
            ("output", {"parts": [{"type": "stdout", "text": "\n4\n"}]}),
        ],
    )


def test_flush_big_output():
    check_simple(
        dedent(
            """
            for i in range(3000):
                print(9)
            """
        ),
        [
            ("output", {"parts": [{"type": "stdout", "text": "9\n" * 500}]}),
        ] * 6,
    )


def test_console_locals():
    runner = MyRunner(callback=default_callback)
    base_locals = {
        "__name__": "__main__",
        "__doc__": None,
        "__package__": None,
        "__loader__": None,
        "__spec__": None,
        "__file__": "my_program.py",
        "__builtins__": builtins.__dict__,
    }

    check_simple("x = 1", [], runner=runner)
    assert runner.console.locals == base_locals | {"x": 1}

    check_simple("y = 2", [], runner=runner)
    assert runner.console.locals == base_locals | {"y": 2}

    check_simple("z = 3", [], runner=runner, mode="single")
    assert runner.console.locals == base_locals | {"y": 2, "z": 3}


def test_await_syntax_error():
    check_simple(
        "await b",
        [
            (
                "output",
                {
                    "parts": [
                        {
                            "type": "syntax_error",
                            "text": (
                                '  File "my_program.py", line 1\n'
                                "SyntaxError: 'await' outside function\n"
                            ),
                            "source_code": "await b",
                        }
                    ]
                },
            )
        ],
    )


def test_sleep():
    class SleepRunner(MyRunner, PatchedSleepRunner):
        pass

    check_simple(
        "import time; time.sleep(123)",
        [
            (
                "sleep",
                {"seconds": 123},
            ),
        ],
        runner=SleepRunner(callback=default_callback),
    )


def test_interrupt():
    with pytest.raises(KeyboardInterrupt):
        check_simple("raise KeyboardInterrupt", [])


def test_await():
    global events
    events = []

    runner = MyRunner(callback=default_callback)
    result = runner.run_async(
        dedent(
            """
    async def foo():
        print('hi')
    
    await foo()
            """
        )
    )
    assert events == []
    asyncio.run(result, debug=True)
    assert events == [
        (
            "output",
            {
                "parts": [
                    {"type": "stdout", "text": "hi\n"},
                ],
            },
        ),
    ]


def test_async_without_await():
    global events
    events = []

    runner = MyRunner(callback=default_callback)
    result = runner.run_async(
        dedent(
            """
    def foo():
        print('hi')

    foo()
            """
        )
    )
    assert events == []
    asyncio.run(result, debug=True)
    assert events == [
        (
            "output",
            {
                "parts": [
                    {"type": "stdout", "text": "hi\n"},
                ],
            },
        ),
    ]


def test_invalid_sleep():
    runner = PatchedSleepRunner()
    for arg in [-1, float('nan'), "", 1+2j, None]:
        with pytest.raises(Exception):
            try:
                time.sleep(arg)
            except Exception as e:
                assert isinstance(e, (ValueError, TypeError))
                with pytest.raises(type(e)):
                    runner.sleep(arg)
                raise



from .runner import Runner, PatchedStdinRunner, PatchedSleepRunner

try:
    from .version import __version__
except ImportError:  # pragma: no cover
    # version.py is auto-generated with the git tag when building
    __version__ = "???"
